/*
 *  linux/arch/arm/mach-ks8695/fasttimer.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/fast_timer.h>

#ifdef CONFIG_MTD_SNAPARM

/*
 * Where is the timer (VA)?
 */
#define TIMER_VA_BASE  (IO_ADDRESS(KS8695_IO_BASE))

/*
 * How long is the timer interval?
 */
#define TIMER_INTERVAL	     (TICKS_PER_uSEC * 2000)
#define TIMER_DATA_VALUE     (TIMER_INTERVAL >> 1)
#define TIMER_PULSE_VALUE    (TIMER_INTERVAL - TIMER_DATA_VALUE)
#define TIMER_VALUE          (TIMER_DATA_VALUE + TIMER_PULSE_VALUE)

static struct timer_list fast_timerlist;

struct ftentry {
	void	(*func)(void *);
	void	*arg;
};

#define FAST_TIMER_MAX  8
static struct ftentry	fast_timer[FAST_TIMER_MAX];
static int		fast_timers = 0;
static spinlock_t	fast_timer_lock;

static void fast_timer_irq(int irq, void *dev_id, struct pt_regs *regs)
{
	int i;

        __raw_writel(KS8695_INTMASK_TIMERINT0, TIMER_VA_BASE + KS8695_INT_STATUS);

	for (i = 0; i < fast_timers; i++)
		(*fast_timer[i].func)(fast_timer[i].arg);
}

static struct irqaction fasttimer_irq  = {
	fast_timer_irq, SA_INTERRUPT, 0, "fast timer", NULL, NULL
};
                                                                                
void fast_timer_setup(int rate)
{
	printk("FAST-TIMER: fast timer %dHz on IRQ %d\n", rate, KS8695_INT_TIMERINT0);

	setup_arm_irq(KS8695_INT_TIMERINT0, &fasttimer_irq);

	/* Setup TIMER0 as fast clock */
        __raw_writel(TIMER_DATA_VALUE, TIMER_VA_BASE + KS8695_TIMER0);
        __raw_writel(TIMER_PULSE_VALUE, TIMER_VA_BASE + KS8695_TIMER0_PCOUNT);
printk("%s(%d): data_value=%d pulse_value=%d\n", __FILE__, __LINE__, TIMER_DATA_VALUE, TIMER_PULSE_VALUE);
        __raw_writel(0x03, TIMER_VA_BASE + KS8695_TIMER_CTRL);
}

void fast_timer_add(void (*func)(void *arg), void *arg)
{
	unsigned long flags;
	int i;

	for (i = 0; i < fast_timers; i++) {
		if (fast_timer[i].func == func && fast_timer[i].arg == arg) {
			printk(KERN_ERR
				"SnapGear: fast timer entry already exists (0x%x, 0x%x)\n",
				(unsigned int) func, (unsigned int) arg);
			return;
		}
	}

	spin_lock_irqsave(&fast_timer_lock, flags);
	if (fast_timers < FAST_TIMER_MAX) {
		fast_timer[fast_timers].func = func;
		fast_timer[fast_timers].arg = arg;
		fast_timers++;
	} else {
		printk(KERN_ERR "SnapGear: fast timer, no free slots\n");
	}
	spin_unlock_irqrestore(&fast_timer_lock, flags);
}

void fast_timer_remove(void (*func)(void *arg), void *arg)
{
	unsigned long flags;
	int i;

	spin_lock_irqsave(&fast_timer_lock, flags);
	for (i = 0; i < fast_timers; i++) {
		if (fast_timer[i].func == func && fast_timer[i].arg == arg) {
			memmove(&fast_timer[i], &fast_timer[i+1],
				sizeof(struct ftentry) * (FAST_TIMER_MAX - (i+1)));
			fast_timers--;
			spin_unlock_irqrestore(&fast_timer_lock, flags);
			return;
		}
	}
	spin_unlock_irqrestore(&fast_timer_lock, flags);
	printk(KERN_ERR "SnapGear: fast timer entry does not exist (0x%x, 0x%x)\n",
		(unsigned int) func, (unsigned int) arg);
}


/*
 *	Return the current ticks on the fast timer.
 */
unsigned long
fast_timer_count(void)
{
	return(0);
}

#endif /* CONFIG_MTD_SNAPARM */
