/*
 *  linux/drivers/ide/pci/via6421.c	Version 0.1	February, 3rd 2005
 *
 *  Copyright (C) 2005, Freecom Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * For the avoidance of doubt the "preferred form" of this code is one which
 * is in an open non patent encumbered format. Where cryptographic key signing
 * forms part of the process of creating an executable the information
 * including keys needed to generate an equivalently functional executable
 * are deemed to be part of the source code.
 */

#undef REALLY_SLOW_IO		/* most systems can safely undef this */

#include <linux/config.h> /* for CONFIG_BLK_DEV_IDEPCI */
#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/pci.h>
#include <linux/ide.h>
#include <linux/init.h>

#include <asm/io.h>

#include "via6421.h"

static unsigned int __init init_via6421_chipset (struct pci_dev *dev, const char *name)
{
	pci_write_config_byte(dev, 0xab, 0x00); // Set PIO Timing fastest
	pci_write_config_byte(dev,0xa0,0x03);
	pci_write_config_byte(dev,0xb4,0x34);
	pci_write_config_byte(dev, 0x40, 0x03); // Enable SATA primary channel
	pci_write_config_byte(dev, 0x41, 0x03); // Enable IRQ for SATA primary channel
	pci_write_config_byte(dev, 0x42, 0x31); // Enable native mode for SATA primary channel
	//pci_write_config_byte(dev, 0x46, 0x00); // force SATA master

	//pci_write_config_byte(dev, 0x45, 0xac); // disable DMA flush data

	return 0;
}

static void via6421_tune_proc(ide_drive_t *drv, u8 mode)
{
	printk(KERN_ERR ":::::::::ENTERING:::::::: via6421 tune proc mode=0x%02x\n",mode);

}


/* Set the speed of the PATA part only.
 * The PATA part is the 3rd device
 */
static int set_via6421_speed(ide_drive_t *drive, u8 speed)
{
	ide_hwif_t *hwif = drive->hwif;
	struct pci_dev	*dev = hwif->pci_dev;


	if(drive->name[2] >'a'){
		printk(KERN_ERR "VIA6421: unable to set speed for SATA drives\n");
		return 1;
	}

	pci_write_config_byte(dev, 0xab, 0x00); // Set PIO Timing fastest

	switch(speed){
		/* The values below for the UDMA modes are estimated only */
		case XFER_UDMA_7:
			pci_write_config_byte(dev, 0xb3, 0xf0); // UDMA 0T cycletime
			break;
		case XFER_UDMA_6:
			pci_write_config_byte(dev, 0xb3, 0xf2); // UDMA 2T cycletime
			break;
		case XFER_UDMA_5:
			pci_write_config_byte(dev, 0xb3, 0xf4); // UDMA 4T cycletime
			break;
		case XFER_UDMA_4:
			pci_write_config_byte(dev, 0xb3, 0xf6); // UDMA 6T cycletime
			break;
		case XFER_UDMA_3:
			pci_write_config_byte(dev, 0xb3, 0xf8); // UDMA 8T cycletime
			break;
		case XFER_UDMA_2:
			pci_write_config_byte(dev, 0xb3, 0xfa); // UDMA 10T cycletime
			break;
		case XFER_UDMA_1:
			pci_write_config_byte(dev, 0xb3, 0xfc); // UDMA 12T cycletime
			break;
		case XFER_UDMA_0:
			pci_write_config_byte(dev, 0xb3, 0xfe); // UDMA 14T cycletime
			break;
		default:
			pci_write_config_byte(dev, 0xb3, 0x0f); // switch off UDMA
			break;
	}
	return(ide_config_drive_speed(drive,speed));
}

static void __init init_via6421_hwif (ide_hwif_t *hwif)
{
	int i;
	u8 tmp8;
	struct pci_dev	*dev = hwif->pci_dev;
	unsigned long *scr_control=(unsigned long *)hwif->sata_scr[SATA_CONTROL_OFFSET];
	unsigned long *scr_status =(unsigned long *)hwif->sata_scr[SATA_STATUS_OFFSET];

	hwif->speedproc = &set_via6421_speed;
	hwif->tuneproc  = &via6421_tune_proc;

	//pci_write_config_byte(dev, 0xb3, 0xff); // enable UDMA cycle time 17T

	if (!(hwif->dma_base))
		return;

	/* Taken from generic.c without trying to understand */
	hwif->atapi_dma = 1;
	hwif->ultra_mask = 0x7f;
	hwif->mwdma_mask = 0x00;
	hwif->swdma_mask = 0x00;
	hwif->udma_four = 1; /* for VIA6421 */

	if (!noautodma)
		hwif->autodma = 1;

	for(i=0;i<2;i++){
		hwif->drives[i].autodma = hwif->autodma;
	}

	if((hwif->name[3]=='1')||(hwif->name[3]=='2')){
		hwif->sata=1;
		pci_write_config_byte(hwif->pci_dev, 0x40, 0x03); // Enable SATA primary channel
		pci_write_config_byte(hwif->pci_dev, 0x41, 0x03); // Enable IRQ for SATA primary channel
		pci_write_config_byte(hwif->pci_dev, 0x42, 0xF1); // Enable native mode for SATA primary channel
		/*
		if(scr_control){
			*scr_control=0x300;
			udelay(1000);
			*scr_control=0x301;
			udelay(1000);
			if((*scr_status&3)==0){
				printk(KERN_ERR "there seems to be no drive on this port!\n");
			}
			if((*scr_status&3)==1){
				printk(KERN_ERR "there is a drive but phy communication is not up\n");
			}
			if((*scr_status&3)==3){
				printk(KERN_ERR "there is a drive and phy communication\n");
			}
		}
		*/
	}
	/* don't probe slave drives on SATA  */
	/*
	if(hwif->channel <2){
		hwif->drives[1].noprobe=1;
	}
	*/

}

static void init_via6421_dma (ide_hwif_t *hwif, unsigned long dmabase)
{
	ide_setup_dma(hwif, dmabase, 8);
}

static int __devinit via6421_init_one(struct pci_dev *dev, const struct pci_device_id *id)
{
	ide_pci_device_t *d = &via6421_chipset[id->driver_data];
	u16 command;
//	unsigned long *ulptr;



	//ulptr=0;
//	while(ulptr<128){
//	    printk("%08X=%08X\n",(ulong)ulptr,*ulptr);
//	    ulptr++;
//	}


	if (dev->device != d->device)
		BUG();

	pci_read_config_word(dev, PCI_COMMAND, &command);
	if(!(command & PCI_COMMAND_IO))
	{
		printk(KERN_INFO "Skipping disabled %s IDE controller.\n", d->name);
		return 1;
	}

	ide_setup_pci_device(dev, d);
	MOD_INC_USE_COUNT;
	return 0;
}

static void via6421_remove(struct pci_dev *dev) /* Device removed (hot unplug)*/
{
	ide_hwif_t *hwif = pci_get_drvdata(dev);

	if (hwif) {
		ide_unregister(hwif->index);
		MOD_DEC_USE_COUNT;
	}
	pci_disable_device(dev);
	/*
	struct ata_host_set *host_set = pci_get_drvdata(pdev);
	struct ata_port *ap;
	unsigned int i;
	Scsi_Host_Template *sht;
	int rc;
	*/

	/*
		Some things to remember
		-- only the drive is removed and has to be probed again when hotplugged
		-- Module is not unregistered because there is at least one device attached to this device
		-- IRQ is still used by other drives on this device




	*/
	/************ STUFF from scsi driver
	ap = host_set->ports[0];
	sht = ap->host->hostt;
	rc = scsi_unregister_module(MODULE_SCSI_HA, sht);

	free_irq(host_set->irq, host_set);
	if (host_set->mmio_base)
		iounmap(host_set->mmio_base);
	if (host_set->ports[0]->ops->host_stop)
		host_set->ports[0]->ops->host_stop(host_set);

	pci_release_regions(pdev);

	for (i = 0; i < host_set->n_ports; i++) {
		struct ata_ioports *ioaddr;

		ap = host_set->ports[i];
		ioaddr = &ap->ioaddr;

		if ((ap->flags & ATA_FLAG_NO_LEGACY) == 0) {
			if (ioaddr->cmd_addr == 0x1f0)
				release_region(0x1f0, 8);
			else if (ioaddr->cmd_addr == 0x170)
				release_region(0x170, 8);
		}
	}

	kfree(host_set);
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
	*/

}

static int via6421_save_state(struct pci_dev *dev, u32 state)    /* Save Device Context */
{
	return 0;
}

static int via6421_suspend(struct pci_dev *dev, u32 state)	/* Device suspended */
{
	return 0;
}
static int via6421_resume(struct pci_dev *dev)	                /* Device woken up */
{
	return 0;
}

static int via6421_enable_wake(struct pci_dev *dev, u32 state, int enable)   /* Enable wake event */
{
	return 0;
}

/* At this time the driver support only one VIA device */
static struct pci_device_id via6421_pci_tbl[] __devinitdata = {
	{ PCI_VENDOR_ID_VIA, PCI_DEVICE_ID_VIA_6421, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
 	{ 0, },
};

static struct pci_driver driver = {
	.name		= "VIA6421 PCI PATA-SATA",
	.id_table	= via6421_pci_tbl,
	.probe		= via6421_init_one,
	/* .remove = via6421_remove, */
	/* .save_state = via6421_save_state, */
	/* .suspend = via6421_suspend, */
	/* .resume = via6421_resume, */
	/* .enable_wake = via6421_enable_wake, */
};

static int via6421_ide_init(void)
{
	return ide_pci_register_driver(&driver);
}

static void via6421_ide_exit(void)
{
	ide_pci_unregister_driver(&driver);
}

module_init(via6421_ide_init);
module_exit(via6421_ide_exit);

MODULE_AUTHOR("Michael Lawatsch");
MODULE_DESCRIPTION("PCI driver module for VIA6421 PCI PATA-SATA");
MODULE_LICENSE("GPL");

EXPORT_NO_SYMBOLS;

